<?php

namespace Drupal\elastic_apm\EventSubscriber;

use Drupal\elastic_apm\ApiServiceInterface;

use Drupal\Core\Path\PathMatcherInterface;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;

use Elastic\Apm\ElasticApm;
use Elastic\Apm\TransactionInterface;
use Psr\Log\LoggerInterface;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\PostResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * The ElasticApm request event subscriber class.
 *
 * @package Drupal\elastic_apm\EventSubscriber
 */
class RequestSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * The Elastic APM service object.
   *
   * @var \Drupal\elastic_apm\ApiServiceInterface
   */
  protected $apiService;

  /**
   * The current path.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;

  /**
   * Logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * A flag whether the master request was processed.
   *
   * @var bool
   */
  protected $processedMasterRequest = FALSE;

  /**
   * The path matcher.
   *
   * @var \Drupal\Core\Path\PathMatcherInterface
   */
  private $pathMatcher;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Constructs a RequestSubscriber object.
   *
   * @param \Drupal\elastic_apm\ApiServiceInterface $api_service
   *   The Elastic APM service object.
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The current route.
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger service.
   * @param \Drupal\Core\Path\PathMatcherInterface $path_matcher
   *   The path matcher.
   * @param \Drupal\Core\Session\AccountProxyInterface $account_proxy
   */
  public function __construct(
    ApiServiceInterface $api_service,
    RouteMatchInterface $route_match,
    LoggerInterface $logger,
    PathMatcherInterface $path_matcher,
    AccountProxyInterface $account_proxy
  ) {
    $this->apiService = $api_service;
    $this->routeMatch = $route_match;
    $this->logger = $logger;
    $this->pathMatcher = $path_matcher;
    $this->currentUser = $account_proxy->getAccount();
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return [
      KernelEvents::TERMINATE => [['onTerminateRequest', 300]],
    ];
  }

  /**
   * End the transaction and send to PHP Agent whenever this event is triggered.
   *
   * @param \Symfony\Component\HttpKernel\Event\PostResponseEvent $event
   *   The event to process.
   */
  public function onTerminateRequest(PostResponseEvent $event) {
    if (!$this->apiService->isPhpAgentEnabled()) {
      return;
    }
    if (!$this->apiService->shouldMonitor()) {
      return;
    }

    $transaction = ElasticApm::getCurrentTransaction();
    $this->setTransactionName($transaction);
    $this->setTransactionLabels($transaction);
  }

  /**
   * Sets the transaction name to be the route name.
   *
   * No change is made if the route match does not return a route name i.e. the
   * name given by the PHP extension will be kept.
   *
   * @param \Elastic\Apm\ElasticApm\TransactionInterface $transaction
   *   The transaction to set the name for.
   */
  protected function setTransactionName(TransactionInterface $transaction) {
    $route_name = $this->routeMatch->getRouteName();
    if ($route_name) {
      $transaction->setName($route_name);
    }
  }

  /**
   * Sets the transaction labels as determined by configuration.
   *
   * @param \Elastic\Apm\ElasticApm\TransactionInterface $transaction
   *   The transaction to set the labels for.
   */
  protected function setTransactionLabels(TransactionInterface $transaction) {
    $tags = $this->prepareAgentOptions();
    if (empty($tags['tags'])) {
      return;
    }

    $context = $transaction->context();
    foreach ($tags['tags'] as $key => $value) {
      $context->setLabel($key, $value);
    }
  }

  /**
   * Add options to pass to the PHP Agents.
   *
   * Currently we are just tagging admin pages.
   *
   * @return array
   *   An array of options to pass to the PHP Agent, such as tags.
   */
  protected function prepareAgentOptions() {
    $tags = [];

    $route_options = $this->routeMatch->getRouteObject()->getOptions();
    if (isset($route_options['_admin_route'])) {
      $tags['route-is-admin'] = TRUE;
    }

    // Add tags based on the path/route patterns configured.
    $tags = $tags + $this->preparePathPatternTags();
    $tags = $tags + $this->prepareRoutePatternTags();

    // Add tags based on whether the user is authenticated or anonymous.
    $tags = $tags + $this->prepareUserRoleTags();

    return ['tags' => $tags];
  }

  /**
   * Provide tags based on the path patterns configured.
   *
   * @return array
   *   An array of tags to pass to the PHP Agent.
   */
  protected function preparePathPatternTags() {
    $tags = [];

    // Fetch the current path from route object.
    $path = $this->routeMatch->getRouteObject()->getPath();

    // Fetch the configured path patterns.
    $tag_config = $this->apiService->getTagConfig();

    if (!$tag_config['path_patterns']) {
      return [];
    }

    $patterns = $this->apiService
      ->parseTagPatterns($tag_config['path_patterns']);

    // Add tags depending on the path pattern set.
    foreach ($patterns as $pattern) {
      // If the configured path does not match with the current path
      // continue to look for the next pattern.
      if (!$this->pathMatcher->matchPath($path, $pattern['pattern'])) {
        continue;
      }

      $tags[$pattern['tag_key']] = $pattern['tag_value'];
    }

    return $tags;
  }

  /**
   * Provide tags based on the route patterns configured.
   *
   * @return array
   *   An array of tags to pass to the PHP Agent.
   */
  protected function prepareRoutePatternTags() {
    $tags = [];

    // Fetch the current path from route object.
    $route = $this->routeMatch->getRouteName();

    // Fetch the configured path patterns.
    $tag_config = $this->apiService->getTagConfig();

    if (!$tag_config['route_patterns']) {
      return [];
    }

    $patterns = $this->apiService
      ->parseTagPatterns($tag_config['route_patterns']);

    // Add tags depending on the path pattern set.
    foreach ($patterns as $pattern) {
      // If the configured route does not match with the current route
      // continue to look for the next pattern.
      if (!$this->matchRoute($route, $pattern['pattern'])) {
        continue;
      }

      $tags[$pattern['tag_key']] = $pattern['tag_value'];
    }

    return $tags;
  }

  /**
   * Provide tags based on whether the user is authenticated/anonymous.
   *
   * @return array
   *   An array of tags to pass to Elastic APM.
   */
  protected function prepareUserRoleTags() {
    $tag_config = $this->apiService->getTagConfig();

    if (empty($tag_config['user_role'])) {
      return [];
    }

    return [
      'user-is-anonymous' => $this->currentUser->isAnonymous(),
      'user-is-authenticated' => $this->currentUser->isAuthenticated(),
    ];
  }

  /**
   * Checks if a route matches any pattern in a set of patterns.
   *
   * @param string $route
   *   The route to match.
   * @param string $pattern
   *   The pattern string.
   *
   * @return bool
   *   TRUE if the route matches the pattern, FALSE otherwise.
   */
  protected function matchRoute($route, $pattern) {
    if (!isset($this->regexes[$pattern])) {
      // Convert path settings to a regular expression.
      $to_replace = [
        // Quote asterisks.
        '/\\\\\*/',
      ];
      $replacements = [
        '.*',
      ];
      $patterns_quoted = preg_quote($pattern, '/');
      $this->regexes[$pattern] = '/^(' . preg_replace($to_replace, $replacements, $patterns_quoted) . ')$/';
    }

    return (bool) preg_match($this->regexes[$pattern], $route);
  }

}
